import { Widget } from "../../../../pi/widget/widget";
import { popNew } from '../../../../pi/ui/root';
import { iterDb, tables, metaInfo } from "../../net/rpc";
import { BonBuffer } from '../../../../pi/util/bon';
import { read as bonRead } from '../../../../pi/struct/util';

import * as MetaInit from "../../../../pi/struct/meta_init";
import { TabMeta } from "../../../../pi/struct/sinfo";

MetaInit.init();
let structMgr = MetaInit.mgr;

declare var pi_modules;

interface Props {
    recentKey: any;
    tables: string[][];
    startKey: any;
    endKey: any;
}

export class App extends Widget{
    public tables: any[];
    public metas: any[];
    public props: Props;

    constructor() {
        super();
        this.props = {
            recentKey: "",
            tables: [],
            startKey: "",
            endKey: "",
        }
        this.tables = [];
        this.metas = [];
    }

    public tabs(ware: string) {
        tables(ware, r => {
            let bon1 = new BonBuffer(r.msg);
            let values = bonRead(bon1, structMgr);
            let idx = values.indexOf("_$sinfo");
            if (idx > -1) {
                values.splice(idx, 1)
            }
            for(let i = 0; i < values.length; i++) {
                this.props.tables[i] = [values[i], ""]
                this.tables[i] = values[i];

                metaInfo(ware, this.tables[i], r => {
                    let bon = new BonBuffer(r.msg);
                    let meta = new TabMeta();
                    meta.bonDecode(bon);
                    this.metas.push(meta)
                    this.props.tables[i][1] = JSON.stringify(meta)
                    this.paint()
                })
            }            
        });
    }

    public firstPaint() {
        super.firstPaint();
        setTimeout(()=>{this.getAllTables()}, 1000);
    }

    public getAllTables() {
        this.tabs("file");
    }

    public openTab(e: any) {
        let tabName = this.tables[e.index];

        iterDb("file", tabName, 10, undefined, undefined, true, r => {
            console.log("code", r.code);
            let [modName, expt] = tabName.split(".");
            let t = pi_modules[modName + ".s"].exports[expt];
            let primaryKey = t._$info.notes.get("primary");

            let bon1 = new BonBuffer(r.msg);
            let values = bon1.readArray(() => {
                let bin = bon1.readBin();
                // undefined 序列化为 0
                if(equal(bin.buffer, new Uint8Array([0]).buffer)) {
                    return null
                }
                // 解析数据表
                let b = new BonBuffer(bin);
                let m = new t();
                m.bonDecode(b);
                return m;
            });

            if (values.length) {
                this.props.recentKey = values[values.length - 1][primaryKey]
            }
            let datas = [];
            for(let v of values) {
                datas.push([v[primaryKey], JSON.stringify(v)]);
            }

            if(r.code === 1) {
                popNew("management-client-app-base-dbtable", {primaryKey: primaryKey, startKey: this.props.startKey, recentKey: this.props.recentKey, datas: datas, title:["key","value","操作"], showBoder:true, tab: this.props.tables[e.index][0]});
            } else {
                popNew("pi-components-message-message", { type: "fail", content: values, center: true });
            }

        });
    }

    public inputStartKey(e: any) {
        console.log("inputStartKey", e);
        this.props.startKey = e.value;
    }

    public inputEndKey(e: any) {
        console.log("inputEndKey" ,e);
        this.props.endKey = e.value;
    }

    public fetchRecords(e: any) {
        console.log("app-fetch: ", this.props.startKey, this.props.endKey)
        console.log("fetchRecords:", e);
    }

    public rangeQuery(e: any) {
        popNew("management-client-app-base-range", {showBoder:true, tables: this.props.tables, index: e.index});
        
        console.log("rangeQuery1212: ", e);
        // console.log("startKey: ", "endKey", this.state.startKey, this.state.endKey);
    }

    public insertData(e: any) {
        console.log("insert data: ", e);
        popNew("management-client-app-base-insert", {showBoder:true, tables: this.props.tables, index: e.index});
    }
}

// 检查是不是undefined序列化的
export const equal = (buf1, buf2) => {
    if (buf1.byteLength != buf2.byteLength) return false;
    var dv1 = new Uint8Array(buf1);
    var dv2 = new Uint8Array(buf2);
    for (var i = 0 ; i != buf1.byteLength ; i++)
    {
        if (dv1[i] != dv2[i]) return false;
    }
    return true;
}