/**
 * 数据库表格
 */

import * as MetaInit from '../../../../pi/struct/meta_init';
import { read as bonRead } from '../../../../pi/struct/util';
import { BonBuffer } from '../../../../pi/util/bon';
import { Widget } from '../../../../pi/widget/widget';

import { NewType } from '../../../../pi/compile/gendrust';
import { TabMeta } from '../../../../pi/struct/sinfo';
import { popNew } from '../../../../pi/ui/root';
import { notify } from '../../../../pi/widget/event';
import { Forelet } from '../../../../pi/widget/forelet';
import { newPrimaryKeyType, newValueFromJsonString } from '../../../utils/util';
import { iterDb, metaInfo, tables, wares, writeDb } from '../../net/rpc';
import { register } from '../../store/memstore';
import { equal } from './app';
MetaInit.init();
const structMgr = MetaInit.mgr;
export const forelet = new Forelet();
declare var pi_modules;
// tslint:disable-next-line:no-reserved-keywords
declare var module: any;
export const WIDGET_NAME = module.id.replace(/\//g, '-');
interface Props {
    showTitleList:any[];// 表格标题
    showDataList:any[];// 表格值
    allCheck:boolean;// 是否全选 1全选 0全不选
    searchString:string;// 搜索的字段
    database:any[];// 数据库列表
    currentTable:any[];// 表
    currentDatabaseName:string;// 当前数据库
    currentTableName:string;// 当前数据表
    currentIndex:number;// 当前页数
    primary:any;// 主键名
    add:number;// 是否添加数据
}

// tslint:disable-next-line:completed-docs
export class Table extends Widget {
    public ok:() => void;
    public props:Props = {
        showTitleList:[],
        showDataList:[],
        allCheck :false,
        database:[],
        currentTable:[],
        currentIndex:0,
        add:-1
    };
    public init() {
        wares(r => {
            console.log('getallwares: ', r);
            const bon = new BonBuffer(r.msg);
            const values = bonRead(bon, structMgr);
            this.props.database = values;
            this.props.currentDatabaseName = values[1];
            tables(this.props.database[1], r => {
                console.log('get tabs: ', r);
                const bon1 = new BonBuffer(r.msg);
                const value = bonRead(bon1, structMgr);
                this.props.currentTable = value;
                this.props.currentTableName = value[1];
                this.getTableTitle();
            });  
        });
    }
    // =======================================================获取表头==========================================================
    public getTableTitle() {
        metaInfo(this.props.currentDatabaseName, this.props.currentTableName, r => {
            console.log('getMetaInfo: ', r);
            const bon = new BonBuffer(r.msg);
    
            const tabMeat = new TabMeta();
    
            tabMeat.bonDecode(bon);
            const data = tabMeat.v.structType.fields;
            // 判断是否存在主键
            if (tabMeat.v.structType.notes) {
                for (const [k, cf] of tabMeat.v.structType.notes) {
                    if (k === 'primary') {
                        this.props.primary = cf;
                    }
        
                }
            } else {
                this.props.primary = undefined;
            }
            this.props.showTitleList = [];
            data.forEach((v) => {
                this.props.showTitleList.push(v.name);
            });
            this.getTableData(100,undefined);
            console.log('==========表头================',tabMeat.v.structType);
        });
    }
    // ==============index:获取数据库前多少个数据========startKey：主键位置================================获取表数据============================================================
    public getTableData(index:number,startKey:any) {
        iterDb(this.props.currentDatabaseName, this.props.currentTableName, index, startKey, null, true,r => {
            console.log('code', r.code);
            const [modName, expt] = this.props.currentTableName.split('.');
            // tslint:disable-next-line:prefer-template
            const t = pi_modules[modName + '.s'].exports[expt];

            if (r.code != 1) {
                console.log("dbiter error");
                return;
            }
    
            const bon1 = new BonBuffer(r.msg);
            const values = bon1.readArray(() => {
                const bin = bon1.readBin();
                // undefined 序列化为 0
                if (equal(bin.buffer, new Uint8Array([0]).buffer)) {
                    return null;
                }
                // 解析数据表
                const b = new BonBuffer(bin);
                const m = new t();
                m.bonDecode(b);

                return m;
            });
            this.props.showDataList = values;
            this.paint();
            console.log('+++++++++++++++123++++++++++++++++',values,this.props.showDataList);
        });
    }
    // 触发选中的数据库值
    public databaseChange(e:any) {
        console.log(e.value);
        const ware = e.value;
        this.props.currentDatabaseName = ware;
        tables(ware, r => {
            console.log('get tabs: ', r);
            const bon1 = new BonBuffer(r.msg);
            const values = bonRead(bon1, structMgr);
            this.props.currentTable = values;
            console.log(values);
            this.paint();
        });
    }
    // 触发选中的表值
    public tableChange(e:any) {
        this.props.currentTableName = e.value;
        this.getTableTitle();
        this.paint();
    }
    // 触发搜索框的值
    public evSearch(e:any) {
        this.props.searchString = e.value;
    }
    // 搜索
    public search() {
        if (this.props.searchString) {
            const data = JSON.parse(this.props.searchString);
            const dataList = [];
            // 筛选跟主键值一样的数据
            this.props.showDataList.forEach(v => {
                if (v[this.props.primary] === data) {
                    dataList.push(v);
                }
            });
            this.props.showDataList = dataList;
        } else {
            this.getTableData(100,undefined);
        }
       
        this.paint();
    }
    // 分页触发函数===============================================================
    public changeCurrent(e:any) {
        console.log('分页+++++++++=',e.value); 
    }
    // 触发删除数据函数   type： 0删除  1修改  2新增
    public tableChangeData(e:any) {
        const primary = this.props.primary;// 主键名字
        const index = e.index; // 下标
        const values = this.props.showDataList[index][primary];// 主键值
        if (e.types === 0) {
            // 删除数据
            writeDb(this.props.currentDatabaseName, this.props.currentTableName, [values], [undefined], r => {
                if (r.code === 1) {
                    popNew('pi-components-message-message', { type: 'success', content: '删除成功', center: true });
                    this.props.showDataList.splice(index,1);
                    this.paint();
                } else {
                    popNew('pi-components-message-message', { type: 'success', content: '删除失败', center: true });
                }
            });
        } else if (e.types === 1) {
            // 修改数据
            const [modName, expt] = this.props.currentTableName.split('.');
            // tslint:disable-next-line:prefer-template
            const t = pi_modules[modName + '.s'].exports[expt];
            const primaryKey = t._$info.notes.get('primary');
            const x = newValueFromJsonString(t, JSON.stringify(e.value));
            writeDb(this.props.currentDatabaseName, this.props.currentTableName, [x[primaryKey]], [x], r => {
                if (r && r.code === 1) {
                    popNew('pi-components-message-message', { type: 'success', content: '修改成功', center: true });
                    this.props.showDataList[index] = e.value;
                    this.paint();
                } else {
                    popNew('pi-components-message-message', { type: 'success', content: '修改失败', center: true });
                }
            });
        } else if (e.types === 2) {
            // 新增一条数据

            const tab = this.props.currentTableName;
            const [modName, expt] = tab.split('.');
            // tslint:disable-next-line:prefer-template
            const t = pi_modules[modName + '.s'].exports[expt];
            const key = newPrimaryKeyType(t._$info, e.value[this.props.primary]);
            const value = newValueFromJsonString(t, JSON.stringify(e.value));
    
            // TODO
            writeDb(this.props.currentDatabaseName, tab, [key], [value], r => {
                console.log(r);
                if (r.code === 1) {
                    popNew('pi-components-message-message', { type: 'success', content: '插入成功', center: true });
                } else {
                    popNew('pi-components-message-message', { type: 'fail', content: '插入失败', center: true });
                }
                this.paint();
            });
        }
       
    }
    // 添加数据
    public addDate() {
        const data = {};
        for (const key in this.props.showDataList[0]) {
            data[key] = '';
        }
        this.props.showDataList.unshift(data);
        this.props.add = 0;
        this.paint();
    }
    public popRpc() {
        popNew("management-client-components-console")
    }
}
register('login',(res:any) => {
    const w:any = forelet.getWidget(WIDGET_NAME);
    w && w.init();
});