import { clientRpcFunc } from './init';
import { Login, AddUser, Auth, Passwd } from '../../server/rpc/user.s';
import { login, addUser as aUser, delUser as dUser, modifyAuth as mAuth, changePasswd as cPasswd } from '../../server/rpc/user.p';
import { ResultStr, ResultBin } from '../../utils/result.s';

import { Read, Select, Write, MetaInfo } from '../../server/rpc/db_mgr.s';
import { read, write, select, getAllTables, getTabMetaInfo, getAllWares } from '../../server/rpc/db_mgr.p';

import { ShowLog, UserLog, WriteLog } from '../../server/rpc/log.s'
import { showLog as sLog } from '../../server/rpc/log.p'

import { BonBuffer } from '../../../pi/util/bon';
import { read as bonRead } from '../../../pi/struct/util';

import * as MetaInit from "../../../pi/struct/meta_init";
import { TabMeta } from '../../../pi/struct/sinfo';

import { User } from '../../server/data/db/user.s'
import { Achievement, AdCount, Buzz } from '../../server/data/db/achieve.s';
import { subscribe, unSubscribe } from './init'


declare var pi_modules;

MetaInit.init();
let structMgr = MetaInit.mgr;

export const readDb = (ware: string, tab: string, keys: any, cb: (r: ResultBin) => void) => {
    let bon = new BonBuffer();

    bon.writeArray(keys, el => {
        let b = new BonBuffer();
        (<any>el).bonEncode(b);
        bon.writeBin(b.getBuffer());
    });

    let read_info = new Read();
    read_info.tab = tab;
    read_info.keys = bon.getBuffer();
    read_info.ware = ware;

    clientRpcFunc(read, read_info, (r: ResultBin) => {
        cb(r)
    });
}

// 检查是不是undefined序列化的
function equal (buf1, buf2)
{
    if (buf1.byteLength != buf2.byteLength) return false;
    var dv1 = new Uint8Array(buf1);
    var dv2 = new Uint8Array(buf2);
    for (var i = 0 ; i != buf1.byteLength ; i++)
    {
        if (dv1[i] != dv2[i]) return false;
    }
    return true;
}

(<any>self).dbRead = (ware: string, tab: string, keys: any) => {
    readDb(ware, tab, keys, (r) => {
        console.log("code: ", r.code);
        let [modName, expt] = tab.split(".");
        let t = pi_modules[modName + ".s"].exports[expt];

        let bon1 = new BonBuffer(r.msg);
        let values = bon1.readArray(() => {
            let bin = bon1.readBin();
            // undefined 序列化为 0
            if(equal(bin.buffer, new Uint8Array([0]).buffer)) {
                return null
            }
            // 解析数据表
            let b = new BonBuffer(bin);
            let m = new t();
            m.bonDecode(b);
            return m;
        });

        console.log(values);
    });
}

export const iterDb = (ware: string, tab: string, pagination: number, startKey: any, endKey: any, dir: boolean = true, cb: (r: ResultBin) => void) => {
    let bon1 = new BonBuffer();
    bon1.write(startKey);

    let bon2 = new BonBuffer();
    bon2.write(endKey);

    let iterInfo = new Select();
    iterInfo.tab = tab;
    iterInfo.startKey = bon1.getBuffer();
    iterInfo.endKey = bon2.getBuffer();
    iterInfo.direction = dir;
    iterInfo.pagination = pagination;
    iterInfo.ware = ware;

    clientRpcFunc(select, iterInfo, (r: ResultBin) => {
        cb(r)
    })
}

// 迭代通过 .rs 文件生成的数据表的例子:

// 升序迭代所有记录： dbIter("file", "management/server/data/db/user.User", 10, undefined, undefined, true)
// 降序迭代所有记录：dbIter("file", "management/server/data/db/user.User", 10, undefined, undefined, false)
// 只能用来操作通过 .rs 文件生成的数据表， 因为要使用 pi_modules 里面的信息， 只有通过 .rs 文件生成的数据表才被注册到了 pi_modules
(<any>self).dbIter = (ware: string, tab: string, pagination: number, startKey: any, endKey: any, dir: boolean = true) => {
    iterDb(ware, tab, pagination, startKey, endKey, dir, r => {
        console.log('code', r.code);
        let [modName, expt] = tab.split(".");
        let t = pi_modules[modName + ".s"].exports[expt];

        if (r.code != 1) {
            console.log("dbiter error");
            return;
        }

        let bon1 = new BonBuffer(r.msg);
        let values = bon1.readArray(() => {
            const bin = bon1.readBin();
            // undefined 序列化为 0
            if(equal(bin.buffer, new Uint8Array([0]).buffer)) {
                return null
            }
            // 解析数据表
            const b = new BonBuffer(bin);
            const m = new t();
            m.bonDecode(b);
            return m;
        });

        console.log(values);
    })
}

export const writeDb = (ware: string, tab: string, keys: any, values: any, cb: (r: ResultStr) => void) => {
    let bon1 = new BonBuffer();
    bon1.writeArray(keys, el => {
        const b = new BonBuffer();
        if (el !== undefined && el !== undefined) {
            (<any>el).bonEncode(b);
        } else {
            b.write(el);
        }
        bon1.writeBin(b.getBuffer());
    });

    let bon2 = new BonBuffer();
    bon2.writeArray(values, el => {
        const b = new BonBuffer();
        if (el !== undefined && el !== undefined) {
            (<any>el).bonEncode(b);
        } else {
            b.write(el);
        }
        bon2.writeBin(b.getBuffer());
    });

    const writeInfo = new Write();
    writeInfo.tab = tab;
    writeInfo.key = bon1.getBuffer();
    writeInfo.value = bon2.getBuffer();
    writeInfo.ware = ware;

    clientRpcFunc(write, writeInfo, (r: ResultStr) => {
        cb(r)
    })
} 

(<any>self).dbWrite = (ware: string, tab: string, keys: any, values: any) => {
    writeDb(ware, tab, keys, values, r => {
        console.log("write result: ", r);
    })
}

(<any>self).writeUser = (pwd: string, user: string, uid: number) => {
    const u = new User();
    u.pwd = 'hhhh'
    u.user = 'xxxx';

    const u2 = new User();
    u2.pwd = 'jjj';
    u2.user = 'kkk';

    const tab = 'management/server/data/db/user.User';
    // uid: 是一个数组
    writeDb('file', tab, uid, [u, u2], r => {
        console.log('write result: ', r);
    });
};

export const tables = (ware: string, cb: (r: ResultBin) => void) => {
    clientRpcFunc(getAllTables, ware, (r: ResultBin) => {
        cb(r)
    });
};

(<any>self).getTabs = (ware: string) => {
    tables(ware, r => {
        console.log('get tabs: ', r);
        const bon1 = new BonBuffer(r.msg);
        const values = bonRead(bon1, structMgr);
        console.log(values);
    });
};

export const wares = (cb: (r: ResultBin) => void)  => {
    clientRpcFunc(getAllWares, undefined, (r: ResultBin) => {
        cb(r)
    });
};

(<any>self).allWares = () => {
    wares(r => {
        console.log('get all wares: ', r);
        const bon = new BonBuffer(r.msg);
        const values = bonRead(bon, structMgr);
        console.log(values)
    });
};

export const metaInfo = (ware: string, tab: string, cb: (r: ResultBin) => void) => {
    const meta = new MetaInfo();
    meta.ware = ware;
    meta.tab = tab;

    clientRpcFunc(getTabMetaInfo, meta, (r: ResultBin) => {
        cb(r)
    });
};

(<any>self).getMetaInfo = (ware: string, tab: string) => {
    metaInfo(ware, tab, r => {
        console.log('getMetaInfo: ', r);
        const bon = new BonBuffer(r.msg);

        const tabMeat = new TabMeta();

        tabMeat.bonDecode(bon);

        console.log(tabMeat)

        return tabMeat
    });
};

(<any>self).setUpTestDb = () => {
    const tab1 = 'management/server/data/db/user.User';

    const keys = [];
    const values = [];

    for (let i = 0; i < 10; i++) {
        const u = new User();
        u.pwd = `foo${i}`;
        u.user = `bar${i}`;

        keys.push(i);
        values.push(u);
    }

    writeDb('file', tab1, keys, values, r => {
        console.log('write result: ', r);
    });

    const keys1 = [];
    const values1 = [];

    const tab2 = 'management/server/data/db/achieve.Achievement';
    for (let i = 0; i < 10; i++) {
        const a = new Achievement();
        a.rid = i;
        a.point = i + 1;
        a.target = i + 2;
        a.business = new Map([['foo', [i + 1,i + 2,i + 3]]]);
        a.makeReduce = new Map();
        a.explore = new Map();
        a.got = [];

        const adcount = new AdCount;
        adcount.dailyTimes = 100,
        adcount.lastTime = 200;
        a.foo = [adcount, adcount];
        a.bar = 100;
        a.baz = adcount;
        a.buzz = [Buzz.HELLO, Buzz.WORLD];

        keys1.push(i);
        values1.push(a);
    }

    writeDb('file', tab2, keys1, values1, r => {
        console.log('write achievement result: ', r);
    });
};

// 导出所有rpc路径
export const getAllRpcsPath = () => {
    const rpcPaths = [];
    for (const k in pi_modules) {
        if (k.endsWith(".p")) {
            rpcPaths.push(k)
        }
    }

    return rpcPaths
}

(<any>self).getAllRpcsPath = () => {
    return getAllRpcsPath();
}

// 导出某个rpc用到的rpc使用的结构
export const getRpcPathExports = (rpcPath: string) => {
    const exps = [];
    for (const e in pi_modules[rpcPath].exports) {
        exps.push(e)
    }

    return exps;
}

(<any>self).getRpcPathExports = (rpcPath: string) => {
    return getRpcPathExports(rpcPath);
}

// 订阅数据库
export const subDb = (path: string, tabName: string, cb: Function) => {
    let [modName, expt] = tabName.split(".");
    let t = pi_modules[modName + ".s"].exports[expt];

    subscribe(path, t, cb);
}

(<any>self).subDb = (path: string, tabName: string) => {
    subDb(path, tabName , r => {
        console.log(r)
    })
}

export const unSubDb = (path: string) => {
    unSubscribe(path);
}

(<any>self).unsubDb = (path: string) => {
    unSubDb(path)
}

// ====================================================================
// ====================================================================

// 用户登录
export const loginUser = (userName: string, pwd: string, cb: (r: ResultStr) => void) => {
    let log = new Login();
    log.user = userName
    log.pwd = pwd;

    clientRpcFunc(login, log, (r: ResultStr) => {
        cb(r);
    });
};

(<any>self).login = (userName: string, pwd: string) => {
    loginUser(userName, pwd, (r) => {
        console.log(r);
    });
};

// 添加用户
export const addUser = (userName: string, pwd: string, cb: (r: ResultStr) => void) => {
    let u = new User;
    u.user = userName
    u.pwd = pwd;

    clientRpcFunc(aUser, u, (r: ResultStr) => {
        cb(r)
    })
}

(<any>self).addUser = (userName: string, pwd: string) => {
    addUser(userName, pwd, r => {
        console.log(r)
    })
}

// 删除用户
export const delUser = (userName: string, cb: (r: ResultStr) => void) => {
    clientRpcFunc(dUser, userName, r => {
        cb(r)
    })
}

(<any>self).delUser = (userName: string) => {
    delUser(userName, r => {
        console.log(r)
    })
}

// 修改用户授权信息
export const modifyAuth = (userName: string, auth: Array<number>, cb: (r: ResultStr) => void) => {
    let au = new Auth;
    au.user = userName;
    au.authList = auth;

    clientRpcFunc(mAuth, au, r => {
        cb(r)
    })
}

(<any>self).modifyAuth = (userName: string, auth: Array<number>) => {
    modifyAuth(userName, auth, r => {
        console.log(r);
    })
}

export const changePasswd = (oldPass: string, newPass: string, cb: (r: ResultStr) => void) => {
    let p = new Passwd;
    p.old = oldPass;
    p.new = newPass;

    clientRpcFunc(cPasswd, p, r => {
        cb(r)
    })
}

(<any>self).changePasswd = (oldPass: string, newPass: string) => {
    changePasswd(oldPass, newPass, r => {
        console.log(r)
    })
}

// 查看某一时间段内的日志， 时间戳需要转换为字符串
export const showLog = (startTime: string, endTime: string, cb: (r) => void) => {
    if (!startTime || !endTime) {
        throw new Error("must specify startTime and endTime")
    }

    let log = new ShowLog;
    log.startTime = startTime;
    log.endTime = endTime;

    clientRpcFunc(sLog, log, r => {
        cb(r)
    });
}

(<any>self).showLog = (startTime: string, endTime: string) => {
    showLog(startTime, endTime, r => {
        if (r.code != 1) {
            return;
        }
        
        const bon = new BonBuffer(r.msg);
        let values = bon.readArray(() => {
            let bin = bon.readBin();
            // undefined 序列化为 0
            if(equal(bin.buffer, new Uint8Array([0]).buffer)) {
                return null
            }
            // 解析数据表
            let b = new BonBuffer(bin);
            let m = new WriteLog;
            m.bonDecode(b);
            return m;
        });
        console.log("values", values)
    })
}