import { ResultBin } from '../../utils/result.s';
import { Log } from '../data/db/log.s';
import { ShowLog, WriteLog } from './log.s';
import { Bucket } from '../../utils/db';
import { WARE_NAME } from '../data/constant';
import { Env } from '../../../pi/lang/env';
import { Tr } from '../../../pi/db/mgr';
import { BonBuffer } from '../../../pi/util/bon';
import { iterDb } from '../../../pi_pt/db';
import { arcDerefVec } from '../../../pi_pt/rust/pi_serv/js_base';
import { checkAuth } from './user.r';
import { AuthEnum } from '../data/db/auth.s';

declare var env: Env;

/**
 * 日志功能模块
 */

/**
 * 获取日志
 */
// #[rpc=rpcServer]
export const showLog = (user: ShowLog): ResultBin => {
    const res = new ResultBin;
    let b = new BonBuffer;
    const logs = [];

    if (!checkAuth(AuthEnum.LOG)) {
        res.code = 3; // 用户无权查看log
        return res;
    }

    if (!user.startTime || !user.endTime) {
        res.code = 2; // 必须指定开始和结束时间
        return res;
    }
    env.dbMgr.read((tr: Tr) => {
        let di = iterDb((<any>tr).inner, WARE_NAME, Log._$info.name, user.startTime, true, "");
        while (true) {
            let elem = di.next();

            if (!elem) break;

            let keyBon = new BonBuffer(arcDerefVec(elem[0]).asSliceU8());
            let key = keyBon.read();
            let v = arcDerefVec(elem[1]).asSliceU8();

            if (key < user.endTime) {
                logs.push(v);
            } else {
                break;
            }
        }
    })

    b.writeArray(logs, el => {
        b.writeBin(el);
    });

    res.code = 1;
    res.msg = b.getBuffer();

    return res;
};

/**
 * 写入操作日志
 */
export const writeLog = (log: WriteLog) => {
    if (!log.user) return;
    const writeLog = new Log();

    writeLog.id = log.user;
    writeLog.ip = log.ip;
    writeLog.arg = log.arg;
    writeLog.module = log.module;
    writeLog.time = Date.now().toString();
    writeLog.action = log.action;

    const logBkt = new Bucket(WARE_NAME, Log._$info.name);
    logBkt.put(writeLog.time, writeLog)
};
